import static java.lang.Math.*;


public class atomFunc{
    static vectorFunc v = new vectorFunc();
    
    public static final vector separate(final atom i)        // return randomized separation force adjusted for mass
    {
        double rx = Math.random()-0.5d;
        double ry = Math.random()-0.5d;
        return v.scale(new vector(rx, ry), i.mass);
    }
    
    public static final atom addforce(final atom i, final vector f)    // accumulate force
    {
        atom j = new atom(i);
        j.force = v.add(j.force, f);
        return j;
    }
    
    public static final atom subforce(final atom i, final vector f)    // negative accumulate force
    {
        atom j = new atom(i);
        j.force = v.sub(j.force, f);
        return j;
    }
    
    public static final atom addVel(final atom i, final vector f)    // accumulate force
    {
        atom j = new atom(i);
        j.vel = v.add(j.vel, f);
        return j;
    }
    
    public static final atom subVel(final atom i, final vector f)    // negative accumulate force
    {
        atom j = new atom(i);
        j.vel = v.sub(j.vel, f);
        return j;
    }
    
    public static final vector repulseSQ(final atom i, final atom j, final double restlen, final double distcentSQ, final double damp, final double deep)    // return self repulsion spring+damper forces
    {
        vector link = v.sub(i.pos, j.pos);
        double multSQ = pow(restlen,2) /  distcentSQ;
        double comp = (i.mass*j.mass)/(i.mass+j.mass);
        
        vector spring = v.scale(v.sub(v.scale(link, multSQ), link), comp*deep); // dist=distance surface to surface. (how deep) K=5000
        vector damper = v.scale((v.sub(j.vel, i.vel)), comp*damp);            // D=damping
        return v.add(spring, damper);
    }
    
    public static final vector invRepulseSQ(final atom i, final vector pos, final double restlen, final double distcentSQ, final double damp, final double deep)    // return self repulsion spring+damper forces
    {
        vector link = v.sub(i.pos, pos);
        double multSQ = pow(restlen,2) /  distcentSQ;
        double comp = i.mass;
        
        vector spring = v.scale(v.sub(v.scale(link, multSQ), link), comp*deep); // dist=distance surface to surface. (how deep) K=5000
        vector damper = v.scale( i.vel, comp*damp);            // D=damping
        return v.sub(spring, damper);
    }
    
    
    public static final atom integrate(final atom i, final double dt)        // integrate system
    {
        atom j = new atom(i);
        j.vel = v.add(j.vel, v.scale(v.scale(j.force, j.imass), dt));     // v = (f/m) *dt   or   v = a*dt  or  v = integrated v
        j.pos = v.add(j.pos, v.scale(j.vel, dt));                           // x += vel *dt
        j.force = new vector(0,0);          //zeroforce after integration
        return j;
    }
    
    //stability cheat
    public static final atom terminalforces(final atom i)        // limit maximum forces in atom accumulator
    {
        atom j = new atom(i);
        j.force = v.limit(j.force, 8000000);
        return j;
    }
    
    public static final atom terminalvelocity(final atom i)        // limit maximum velocity
    {
        atom j = new atom(i);
        j.vel = v.limit(j.vel, 10000);
        return j;
    }
    
    
    //spring
    
    public static final atom linkWith(final atom atomSrc, final int atomNum, final double length)
    {
        if (atomSrc.linkQty < atomSrc.maxLink)
        {
            atom nuatom = new atom(atomSrc);
            nuatom.linkedTo[nuatom.linkQty] = atomNum;
            //nuatom.breaklimit[nuatom.linkQty] = 30000;
            //nuatom.rigidity[nuatom.linkQty] = 300;
            nuatom.lenght[nuatom.linkQty] = length;
            nuatom.linkQty++;
            return nuatom;
        }
        return atomSrc;
    }
    
    public static final Boolean linkedWith(final atom atomSrc, final int atomId)
    {
        Boolean test = false;
        for (int i=0;i<atomSrc.linkQty;i++)
            if (atomSrc.linkedTo[i] == atomId)
                test = true;
        return test;
    }
    
    //correct these
    public static final atom springForceSQ(final atom atomSrc, final atom[] atomLst)
    {
        if (atomSrc.linkQty == 0)
            return atomSrc;
        atom nuatom = new atom(atomSrc);
        atom atomDest;
        for (int i=0; i<atomSrc.linkQty;i++)
        {
            atomDest = atomLst[atomSrc.linkedTo[i]];
            //link removal ..... to optimize
            if (!linkedWith(atomDest, atomSrc.id))
                nuatom = unLink(nuatom, i);
            else
            {
                //link not removal
                double distCentersSQ = v.distSQ(atomSrc.pos, atomDest.pos);  // returns SQdistance of atoms centers
                if (distCentersSQ == 0)
                {
                    //distCentersSQ = 0.1d;                     // case distance = 0, make it work. (could delete spring too or break)
                    nuatom = unLink(nuatom, i);
                }
                else
                {
                    vector nuforce = repulseSQ(atomSrc, atomDest, atomSrc.lenght[i], distCentersSQ, atomSrc.rigidity,0.1);
//                    if (v.mag(nuforce)/atomSrc.mass>2500000)
//                        nuatom = unLink(nuatom, i);
//                    else
                        nuatom = addforce(nuatom, nuforce);
                }
            }
        }
        return nuatom;
    }
    
    //not safe yet
    public static final atom unLink(final atom atomSrc, final int linkNum)
    {
        if (atomSrc.linkQty > linkNum)
        {
            atom nuAtom = new atom(atomSrc);
            nuAtom.linkedTo[linkNum] = atomSrc.linkedTo[atomSrc.linkQty-1];
            nuAtom.linkedTo[nuAtom.linkQty-1] = atomSrc.linkedTo[linkNum];
            nuAtom.lenght[linkNum] = atomSrc.lenght[atomSrc.linkQty-1];
            nuAtom.lenght[nuAtom.linkQty-1] = atomSrc.lenght[linkNum];
//            nuAtom.rigidity[linkNum] = atomSrc.rigidity[atomSrc.linkQty];
//            nuAtom.rigidity[nuAtom.linkQty] = atomSrc.rigidity[linkNum];
//            nuAtom.breaklimit[linkNum] = atomSrc.breaklimit[atomSrc.linkQty];
//            nuAtom.breaklimit[nuAtom.linkQty] = atomSrc.breaklimit[linkNum];
            nuAtom.linkQty--;
            return nuAtom;
        }
        return atomSrc;
    }
};
